import functools
import logging
import re

from typing import List, Union, Callable

from azure.graphrbac import GraphRbacManagementClient
from azure.graphrbac.models.graph_error_py3 import GraphErrorException
from msrest.exceptions import AuthenticationError

from ppa import ui

from .._common import service_principal_credentials
from ..exceptions import PermissionDenied, AuthenticationFailed, ObjectAlreadyExists
from ..records import AzureClient

from .records import GroupData, Group, User, UserData


logger = logging.getLogger(__name__)


def api_call(func: Callable):
    @functools.wraps(func)
    def _execute(*args, **kwargs) -> Callable:
        # This wrapper catches exceptions generic to all API calls. More specific errors that require
        # context from an action are left out, & handled by the wrapped function itself.
        try:
            return func(*args, **kwargs)
        except GraphErrorException as e:
            error = str(e).lower()
            if "insufficient privileges" in error:
                message = "The API application does not have permission to perform this operation."
                ui.output_error(message)
                raise PermissionDenied(message)
            elif "already exists" in error:
                message = "The operation failed as a matching object already exists."
                ui.output_error(message)
                raise ObjectAlreadyExists(message)
            raise e
        except AuthenticationError as e:
            message = None
            if error := vars(e.inner_exception).get("error_response", {}).get("error_description"):
                if re.search(re.compile("Tenant '.*' not found"), error):
                    message = "Authentication failed. No tenant was found with the supplied ID."
                elif re.search(re.compile("Application with identifier '.*' was not found"), error):
                    message = "Authentication failed. No application was found with the supplied client ID."
                elif "Invalid client secret is provided" in error:
                    message = "Authentication failed. The supplied client secret is invalid."
            if message:
                ui.output_error(message)
                raise AuthenticationFailed(message)

            ui.output_error("Unhandled authentication failure. Please view the logs for more information.")
            raise e

    return _execute


def _as_iterable(items):
    # Some endpoints return a single object, some return a paging iterator.
    return items if hasattr(items, "__iter__") else [items]


def _create_user(user: dict) -> User:
    return UserData(**user.as_dict()).__dict__


def _create_group(group: dict) -> User:
    return GroupData(**group.as_dict()).__dict__


def process_groups(groups: list) -> List[Group]:
    return list(map(_create_group, _as_iterable(groups)))


def process_users(users: Union[list, dict]) -> List[User]:
    return list(map(_create_user, _as_iterable(users)))


def process_objects(objects: Union[list, dict]) -> List[Union[User, Group]]:
    # This is just a dispatcher for functions that can return a list of both users & groups.
    record_map = {"User": _create_user, "Group": _create_group}

    def _create(obj):
        try:
            return record_map[obj.object_type](obj)
        except KeyError:
            message = f"Object with type '{obj.object_type}' is not supported & will be ignored."
            ui.output_error(message)
            logger.error(message)

    return list(map(lambda o: _create(o), objects))


def process_bool(response: dict) -> bool:
    """
    Example boolean response from an API call:

    {
        'additional_properties': {
            'odata.metadata': 'https://graph.windows.net/84876d51-c3a2-412b-82ac-ba1aa2779357/$metadata#Edm.Boolean'
        },
        'value': True
    }
    """
    try:
        return response.as_dict()["value"]
    except KeyError:
        logger.error(f"response: {response}")
        message = "Failed to process the API response as it does not contain a value. Please view the logs for more information."
        ui.output_error(message)
        raise KeyError(message)


def client(client: AzureClient) -> GraphRbacManagementClient:
    return GraphRbacManagementClient(
        service_principal_credentials(client=client, resource="https://graph.windows.net"), client["tenant_id"]
    )
