"""AD - Users

This module contains actions for viewing and managing Azure AD Users.
"""
import uuid
import logging

from typing import List, Dict, Optional, Any, Union

from azure.graphrbac.models import UserCreateParameters, UserUpdateParameters, PasswordProfile
from azure.graphrbac.models.graph_error_py3 import GraphErrorException

from ppa import ui

from ..exceptions import ObjectNotFound, InvalidPassword
from .._common import get_table_field
from ..records import AzureClient
from ._ad_api import client, api_call, process_users, process_groups
from .records import User, Group, UserUpdateDict


logger = logging.getLogger(__name__)


USER_TABLE_HEADER = ["Display Name", "Principal Name", "Email Address", "Enabled"]
USER_TABLE_FIELDS = ["display_name", "user_principal_name", "mail", "account_enabled"]


@api_call
def get_all(azure_client: AzureClient) -> List[User]:
    """Get all AD users.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

    ??? success "Output"

        A list of [User](../inputs_outputs/#user) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.get_all:
          load:
            azure_client: azure_secrets
          save: all_users
        ```
    """
    return process_users(client(azure_client).users.list())


@api_call
def get_by_name(azure_client: AzureClient, name: str) -> List[User]:
    """Get all users with the supplied display name.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __name__: the user display name

    ??? success "Output"

         A list of [User](../inputs_outputs/#user) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.get_by_name:
            name: Jon Smith
          load:
            azure_client: azure_secrets
          save: user
        ```
    """
    # https://docs.microsoft.com/en-gb/previous-versions/azure/ad/graph/api/entity-and-complex-type-reference#user-entity
    return process_users(client(azure_client).users.list(filter=f"displayName eq '{name}'"))


@api_call
def get_by_principal_name(azure_client: AzureClient, principal_name: str) -> User:
    """Get the user with the supplied user principal name.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __user_principal_name__: the user's principal name

    ??? success "Output"

         A single [User](../inputs_outputs/#user) dictionary.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.get_by_principal_name:
            principal_name: john.smith@domain.com
          load:
            azure_client: azure_secrets
          save: user
        ```
    """
    # https://docs.microsoft.com/en-gb/previous-versions/azure/ad/graph/api/entity-and-complex-type-reference#user-entity
    try:
        return process_users(client(azure_client).users.list(filter=f"userPrincipalName eq '{principal_name}'"))[0]
    except IndexError:
        message = f"No user was found with user principal name '{principal_name}'."
        ui.output_error(message)
        raise ObjectNotFound(message)


@api_call
def get_by_id(azure_client: AzureClient, object_id: str) -> User:
    """Get the user with the supplied object ID.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id__: the user object ID

    ??? success "Output"

         A single [User](../inputs_outputs/#user) dictionary.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.get_by_id:
          load:
            object_id: user_object_id
            azure_client: azure_secrets
          save: user
        ```
    """
    try:
        return process_users(client(azure_client).users.get(object_id))[0]
    except IndexError:
        message = f"No user was found with object ID '{object_id}'."
        ui.output_error(message)
        raise ObjectNotFound(message)


@api_call
def exists(azure_client: AzureClient, principal_name: str) -> bool:
    """Establish whether a user exists with the supplied principal name.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __principal_name__: the principal name to search for

    ??? success "Output"

        A boolean is outputted by this action:

        - `true` if a user exists

        - `false` if no user exists

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.exists:
            principal_name: john.smith@domain.com
          load:
            azure_client: azure_secrets
          save: exists
        ```
    """
    try:
        # Don't call get_by_principal_name as that would output an error in the UI if the user doesn't exist.
        _ = list(client(azure_client).users.list(filter=f"userPrincipalName eq '{principal_name}'"))[0]
        return True
    except IndexError:
        return False


@api_call
def create(
    azure_client: AzureClient,
    display_name: str,
    user_principal_name: str,
    mail_nickname: str,
    password: str,
    mail: Optional[str] = None,
    enabled: Optional[bool] = False,
    force_password_change: Optional[bool] = False,
    additional_properties: Optional[Dict[str, Any]] = None,
) -> User:
    """Create a new user.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __display_name:__ a display name for the user

        - __user_principal_name:__ a user principal name in user@domain format

        - __mail_nickname:__ a mail alias for the user

        - __enabled:__ set to `true` to create the user in an enabled state (defaults to `false`)

        - __password:__ a password to set for the user

        - __force_password_change:__ set to `true` to force password reset on first logon (defaults to `false`)

        - __additional_properties:__ a dictionary containing any other attributes to set (see [here](https://docs.microsoft.com/en-us/python/api/azure-graphrbac/azure.graphrbac.models.usercreateparameters?view=azure-python) for more information)

    ??? success "Output"

         A single [User](../inputs_outputs/#user) dictionary.

    ??? example
        ```YAML linenums="1"
        {% raw %}
        - azure.ad.users.create:
            display_name: John Smith
            user_principal_name: john.smith@domain.com
            mail: john.smith@domain.com
            mail_nickname: john.smith
            enabled: true
            password: "{{ new_password }}"
          load:
            azure_client: azure_secrets
          save: new_user
        {% endraw %}
        ```
    """
    # https://docs.microsoft.com/en-us/python/api/azure-graphrbac/azure.graphrbac.models.usercreateparameters?view=azure-python
    try:
        return process_users(
            client(azure_client).users.create(
                UserCreateParameters(
                    display_name=display_name,
                    user_principal_name=user_principal_name,
                    mail_nickname=mail_nickname,
                    account_enabled=enabled,
                    mail=mail,
                    password_profile=PasswordProfile(password=password, force_change_password_next_login=force_password_change),
                    additional_properties=additional_properties if additional_properties else {},
                )
            )
        )[0]
    except GraphErrorException as e:
        if "does not comply" in str(e).lower():
            message = "Failed to set new password as it does not meet complexity requirements."
            ui.output_error(message)
            raise InvalidPassword(message)
        raise e


@api_call
def set_password(azure_client: AzureClient, object_id: str, password: str, force_password_change: Optional[bool] = False) -> None:
    """Set a user's password.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the user's object ID

        - __password:__ the password to set

        - __force_password_change:__ set to `true` to force password reset on next logon (defaults to `false`)

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        {% raw %}
        ```YAML linenums="1"
        - azure.ad.users.get_by_name:
            name: John Smith
          load:
            azure_client: azure_client
          save: user

        - azure.ad.users.set_password:
            password: "{{ new_password }}"
          load:
            object_id: user.object_id
            azure_client: azure_secrets
          save: new_user
        ```
        {% endraw %}
    """
    try:
        client(azure_client).users.update(
            object_id,
            UserUpdateParameters(
                **{"password_profile": PasswordProfile(password=password, force_change_password_next_login=force_password_change)}
            ),
        )
    except GraphErrorException as e:
        if "does not comply" in str(e).lower():
            message = "Failed to set new password as it does not meet complexity requirements."
            ui.output_error(message)
            raise InvalidPassword(message)
        raise e


@api_call
def delete_by_id(azure_client: AzureClient, object_id: str) -> None:
    """Delete the user with the supplied object ID.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the user's object ID

    !!! warning "Delete Permissions"

        Deleting an object requires your API application in Azure to be assigned the Global Administrators role.

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.get_by_name:
            name: John Smith
          load:
            azure_client: azure_client
          save: user

        - azure.ad.users.delete_by_id:
          load:
            object_id: user.object_id
            azure_client: azure_secrets
        ```
    """
    client(azure_client).users.delete(object_id)


# Not decorated as it calls other decorated functions
def delete_by_principal_name(azure_client: AzureClient, principal_name: str) -> None:
    """Delete the user with the supplied user principal name.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __name:__ the name of the user to delete

    !!! warning "Delete Permissions"

        Deleting an object requires your API application in Azure to be assigned the Global Administrators role.

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.delete_by_principal_name:
            name: John Smith
          load:
            azure_client: azure_secrets
        ```
    """
    client(azure_client).users.delete(get_by_principal_name(azure_client, principal_name)["object_id"])


@api_call
def group_memberships(azure_client: AzureClient, object_id: str) -> List[Group]:
    """Get an all group memberships for an AD user

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the user's object ID

    ??? success "Output"

         A list of [Group](../inputs_outputs/#group) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.group_memberships:
          load:
            object_id: user.object_id
            azure_client: azure_secrets
          save: user_groups
        ```
    """
    return process_groups(
        client(azure_client).groups.list(
            filter="objectID eq '{}'".format(
                "' or objectID eq '".join(
                    [group_id for group_id in client(azure_client).users.get_member_groups(object_id, False)]
                )
            )
        )
    )


@api_call
def update(azure_client: AzureClient, object_id: str, updates: UserUpdateDict) -> None:
    """Update one or more attributes for a user.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __object_id:__ the user object ID

        - __updates__: an attribute update dictionary (see _Update Attributes_ below for more information)

    ??? tip "Update Attributes"

        The following attributes can be supplied __directly__ in the `updates` dictionary:

        - usage_location
        - given_name
        - surname
        - account_enabled
        - display_name
        - user_principal_name
        - mail_nickname

        Attributes _not_ listed above must be supplied inside the following key in the `updates` dictionary:

        - additional_properties

        The example below demonstrates how to do this.

    ??? success "Output"

         Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.ad.users.update:
            updates:
              mail_nickname: john.smith
              additional_properties:
                mail: john.smith@domain.com
          load:
            object_id: user.object_id
            azure_client: azure_secrets
        ```
    """
    client(azure_client).users.update(object_id, UserUpdateParameters(**updates))


def output_table(text: str, users: Union[List[User], User]) -> None:
    """Display a list of users in a table.

    The table will have the following columns:

    - Display Name
    - Principal Name
    - Object ID

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __users:__ a single or list of [User](../inputs_outputs/#user) dictionaries to display in the table

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        - Show a list of users in a table

        ```YAML linenums="1"
        - azure.ad.users.output_table:
            text: Users in Group
          load:
            users: group_users
        ```
    """
    users = [users] if not isinstance(users, list) else users
    ui.output_table(
        text, header=USER_TABLE_HEADER, data=[[get_table_field(user, field) for field in USER_TABLE_FIELDS] for user in users],
    )


def input_table(
    text: str, users: Union[User, List[User]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[User]:
    """Display users in a table, & allow the task operator to make a selection.

    The table will have the following columns:

    - Display Name
    - Principal Name
    - Object ID

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __users:__ a single or list of [User Dictionaries](../inputs_outputs/#user) to display in the table

        - __minimum:__ the minimum number of acceptable selections

        - __maximum:__ the maximum number of acceptable selections

    ??? success "Output"

        A list of [User](../inputs_outputs/#user) dictionaries.

    ??? tip

        - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
        - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
            of selections.

    ??? example

        ```YAML linenums="1"
        - active_directory.users.get_all:
          load:
            azure_client: azure_client
          save: all_users

        - active_directory.users.input_table:
            text: Please select up to 3 users
            minimum: 1
            maximum: 3
          load:
            users: all_users
          save: selected_users
        ```
    """
    users = [users] if not isinstance(users, list) else users
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=USER_TABLE_HEADER,
            data=[[get_table_field(user, field) for field in USER_TABLE_FIELDS] for user in users],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [users[i] for i in indices]
