from typing import TypedDict, Dict, List
from datacaster.classes import CastDataClass


class SecurityGroupRuleData(CastDataClass):
    name: str
    description: str
    protocol: str
    direction: str
    access: str
    priority: int
    source_port_range: str
    source_address_prefix: str
    source_address_prefixes: List[str]
    destination_port_range: str
    destination_address_prefix: str
    destination_address_prefixes: List[str]


class SecurityGroupRule(TypedDict):
    """
    - Contains information about an Azure Network Security Group rule

    !!! info "SecurityGroupRule Keys"
        __Name:__ Rule name

        __description:__ Rule description

        __protocol:__ Protocol (`tcp`, `udp`, or `*` for any)

        __direction:__ Direction (`inbound` or `outbound`)

        __access:__ Access (`allow` or `deny`)

        __priority:__ Rules are processed in priority order; the lower the number, the higher the priority.

        __source_port_range:__ Source port range

        __source_address_prefix:__ Source address

        __source_address_prefixes:__ Source addresses (if multiple)

        __destination_port_range:__ Destination port range

        __destination_address_prefix:__ Destination address

        __destination_address_prefixes:__ Destination addresses (if multiple)
    """

    __annotations__ = SecurityGroupRuleData.__dict__["__annotations__"]


class SecurityGroupData(CastDataClass):
    id: str
    name: str
    location: str
    tags: Dict[str, str] = {}
    security_rules: List[SecurityGroupRule]
    __class_config__ = {
        "cast_functions": {"fields": {"security_rules": lambda rules: [SecurityGroupRuleData(**rule).__dict__ for rule in rules]}}
    }


class SecurityGroup(TypedDict):
    """
    - Contains information about an Azure Network Security Group

    !!! info "SecurityGroup Keys"

        __id__:  Resource ID (including Resource Group)

        __name__:  Resource name

        __location__:  Resource location

        __tags__:  Dictionary of tags
    """

    __annotations__ = SecurityGroupData.__dict__["__annotations__"]
